import crate.client.cursor
from django.db.backends.base.introspection import BaseDatabaseIntrospection, TableInfo


class DatabaseIntrospection(BaseDatabaseIntrospection):
    ignored_tables = []

    def get_table_list(self, cursor: crate.client.cursor.Cursor):
        cursor.execute(
            """
            SELECT
              table_name,
              CASE
                when table_type = 'VIEW' then 'v'
                ELSE 't'
              END
            FROM
              "information_schema"."tables"
            """
        )

        return [
            TableInfo(*row)
            for row in cursor.fetchall()
            if row[0] not in self.ignored_tables
        ]



from django.db.backends.base.creation import BaseDatabaseCreation


class DatabaseCreation(BaseDatabaseCreation):
    def destroy_test_db(
        self, old_database_name=None, verbosity=1, keepdb=False, suffix=None
    ):
        pass

    def _create_test_db(self, *a, **kw):
        pass



from django.db.models.sql.compiler import (
    SQLCompiler,
)  # noqa -- Import needed for re-imports.

from django.db.models.sql.compiler import (
    SQLInsertCompiler,
    SQLUpdateCompiler,
    SQLAggregateCompiler,
    SQLDeleteCompiler,
)


class SQLInsertCompiler(SQLInsertCompiler):
    pass


class SQLDeleteCompiler(SQLDeleteCompiler):
    pass


class SQLUpdateCompiler(SQLUpdateCompiler):
    pass


class SQLAggregateCompiler(SQLAggregateCompiler):
    pass



from django.db.backends.base.features import BaseDatabaseFeatures


class DatabaseFeatures(BaseDatabaseFeatures):
    # Does the backend support partial indexes (CREATE INDEX ... WHERE ...)?
    supports_partial_indexes = False
    supports_functions_in_partial_indexes = False

    # Does the backend support indexes on expressions?
    supports_expression_indexes = False

    supports_foreign_keys = False
    supports_comments = False

    can_rollback_ddl = False
    can_return_columns_from_insert = True

    def supports_transactions(self):
        return False



from django.db import connection

from tests.utils import captured_queries


def test_captured_queries():
    with captured_queries(connection) as ctx:
        connection.cursor().execute("select 1")
        connection.cursor().execute("select 2")
        connection.cursor().execute("select 3")

        assert ctx.first_query.stmt == "select 1"
        assert ctx.query_at(1).stmt == "select 2"
        assert ctx.latest_query.stmt == "select 3"



from django.db.models import JSONField


class ObjectField(JSONField):
    def from_db_value(self, value, expression, connection):
        return value

    def get_internal_type(self):
        return "ObjectField"



import logging
import os

from django.db.backends.base.schema import BaseDatabaseSchemaEditor


class DatabaseSchemaEditor(BaseDatabaseSchemaEditor):
    # TODO DOCUMENT CAVEAT: IF YOU START WITH A DJANGO MIGRATIONS CREATED BY OTHER DATABASE LIKE POSTGRES,
    # NEW MIGRATIONS WITH NO-OP operations like drop constraint, might produce confusing behaviour, you might
    # expect things that will be no-op, like drop constraints.

    # TODO pgdiff
    # TODO, is this bug? why isn't supports_deferrable_unique_constraints respected on table creation?

    sql_create_unique = "select 1"

    sql_alter_column_type = "SELECT 1"
    sql_alter_column_null = "SELECT 2"
    sql_alter_column_not_null = "SELECT 3"
    sql_alter_column_default = "SELECT 4"
    sql_alter_column_no_default = "SELECT 5"

    sql_delete_column = "ALTER TABLE %(table)s DROP COLUMN %(column)s"
    sql_delete_table = "DROP TABLE %(table)s"

    def add_index(self, model, index):
        return None

    def rename_index(self, model, old_index, new_index):
        return None

    def remove_index(self, model, index):
        return None

    def add_constraint(self, model, constraint):
        return None

    def remove_constraint(self, model, constraint):
        return None

    def _model_indexes_sql(self, model):
        """
        todo pgdiff

        This overload stops django from issuing CREATE INDEX statements.
        https://forum.djangoproject.com/t/dont-issue-create-index-on-initial-migration/36227/4
        """
        return ()

    def _alter_column_null_sql(self, model, old_field, new_field):
        return ()

    def column_sql(self, model, field, include_default=False):
        if field.unique:
            # todo pgdiff
            if not os.getenv("SUPPRESS_UNIQUE_CONSTRAINT_WARNING", "false") == "true":
                logging.warning(
                    f"CrateDB does not support unique constraints but `{model}.{field}` is set as"
                    f" unique=True, it will be ignored."
                )
            field.unique = False
        return super().column_sql(
            model=model, field=field, include_default=include_default
        )

    def alter_field(self, model, old_field, new_field, strict=False):
        if old_field.get_internal_type != new_field.get_internal_type:
            # You cannot change types after table creation.
            return
        return super().alter_field(model, old_field, new_field, strict)



from django.db import models, connection
from django.db.models.base import ModelBase

# Tuple of all the extra options a CrateModel Meta class has.
# (name, default_value)
CRATE_META_OPTIONS = (
    ("auto_refresh", False),  # Automatically refresh a table on inserts.
)


class MetaCrate(ModelBase):
    def __new__(cls, name, bases, attrs, **kwargs):
        crate_attrs = {}

        # todo document

        try:
            meta = attrs["Meta"]
            for crate_attr in CRATE_META_OPTIONS:
                attr_name = crate_attr[0]
                if attr_name in meta.__dict__:
                    crate_attrs[attr_name] = meta.__dict__[attr_name]
                    delattr(meta, attr_name)

        except KeyError:
            # Has no meta class
            pass

        o = super().__new__(cls, name, bases, attrs, **kwargs)

        # Return back the crate_attrs we took from meta to the already created object.
        for k, v in crate_attrs.items():
            setattr(o._meta, k, v)
        return o


class CrateModel(models.Model, metaclass=MetaCrate):
    """
    A base class for Django models with extra CrateDB specific functionality,

    Methods:
        refresh: Refreshes the given model (table)
    """

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)  # perform the actual save (insert or update)
        auto_refresh = getattr(self._meta, "auto_refresh", False)
        if auto_refresh and self.pk:  # If self.pk is available, its an insert.
            table_name = self._meta.db_table
            with connection.cursor() as cursor:
                cursor.execute(f"refresh table {table_name}")

    @classmethod
    def refresh(cls):
        with connection.cursor() as cursor:
            cursor.execute(f"refresh table {cls._meta.db_table}")

    class Meta:
        abstract = True



import pprint

from django.forms.models import model_to_dict
from tests.test_app.models import AllFieldsModel, SimpleModel, RefreshModel
from django.db import connection
from django.test.utils import CaptureQueriesContext

from tests.utils import captured_queries


def test_model_refresh():
    """Test that Model.refresh() works"""
    with CaptureQueriesContext(connection) as ctx:
        SimpleModel.refresh()
        assert "refresh table test_app_simplemodel" in ctx.captured_queries[0]["sql"]


def test_model_refresh_meta():
    """Test that a refresh statement is sent after updating or inserting when auto_refresh=True in Meta"""
    with captured_queries(connection) as ctx:
        # Test insert
        RefreshModel.objects.create(field="sometext")
        assert ctx.latest_query.stmt == "refresh table test_app_refreshmodel"


def test_model_auto_pk_value_exists():
    """Test that when we create a model object with Django created 'id', the value gets added to the Object"""
    obj = SimpleModel.objects.create(field="test_model_auto_pk_value_exists")
    SimpleModel.refresh()
    assert obj.id
    assert obj.pk
    assert obj.id == obj.pk
    assert isinstance(obj.id, int)


def test_insert_model_field():
    """Test that we can insert a model and refresh it"""
    assert SimpleModel.objects.count() == 0
    with captured_queries(connection) as ctx:
        SimpleModel.objects.create(field="test_insert_model_field")
        assert 'INSERT INTO "test_app_simplemodel"' in ctx.latest_query.stmt

        SimpleModel.refresh()
        assert SimpleModel.objects.count() == 1


def test_update_model():
    with captured_queries(connection) as ctx:
        obj = SimpleModel.objects.create(field="text")
        pk = obj.pk
        assert obj.field == "text"

        obj.field = "sometext"
        obj.save()



        assert obj.field == "sometext"
        assert pk == obj.pk  # Pk did not change

        # TODO: If we do not refresh here it breaks, there should only be 1 record at this time
        # but there are 2, see issue: https://github.com/surister/cratedb-django/issues/7
        SimpleModel.refresh()
        # assert SimpleModel.objects.count() == 1
        # pprint.pp(ctx.captured_queries)

def test_delete_from_model():
    with captured_queries(connection) as ctx:
        assert SimpleModel.objects.count() == 0

        SimpleModel.objects.create()
        SimpleModel.refresh()

        assert SimpleModel.objects.count() == 1
        SimpleModel.objects.all().delete()
        assert ctx.latest_query.stmt == 'DELETE FROM "test_app_simplemodel"'

        SimpleModel.refresh()
        assert SimpleModel.objects.count() == 0


def test_insert_all_fields():
    """Test that an object is created and accounted for with all supported field types"""

    expected = {
        "id": 29147646,
        "field_int": 1,
        "field_int_unique": 2,
        "field_int_not_indexed": 3,
        "field_int_not_null": 4,
        "field_int_null": 5,
        "field_int_default": 6,
        "field_float": 0.1,
        "field_char": "somechar",
        "field_bool": True,
        # "field_date": datetime.datetime(2025, 4, 22, 0, 0, tzinfo=datetime.timezone.utc),
        # "field_datetime": datetime.datetime(1, 1, 1, 1, 1, 1, 1),
        "field_json": {"hello": "world"},
        "field_uuid": "00bde3702f844402b750c1b37d589084",
    }
    AllFieldsModel.objects.create(**expected)
    AllFieldsModel.refresh()
    assert AllFieldsModel.objects.count() == 1

    obj = AllFieldsModel.objects.get()
    assert model_to_dict(obj) == expected



from django.db.models.expressions import Func

"""
class SomeModel(CrateModel):
    id = models.TextField(primary_key=True, db_default=UUID())
    some = models.TextField()
"""


class UUID(Func):
    """https://cratedb.com/docs/crate/reference/en/latest/general/builtins/scalar-functions.html#gen-random-text-uuid"""

    function = "gen_random_text_uuid"



from django.db import connection, models
from cratedb_django.models import functions


def test_field_with_uuid_default():
    """
    Tests a Model field with a db_default of UUID
    """

    class TestModel(models.Model):
        f = models.TextField(db_default=functions.UUID())

        class Meta:
            app_label = "ignore"

    with connection.schema_editor() as schema_editor:
        sql, params = schema_editor.column_sql(
            TestModel, TestModel._meta.get_field("f")
        )
        assert sql == "text DEFAULT (gen_random_text_uuid()) NOT NULL"



from django.db.backends.base.client import BaseDatabaseClient


class DatabaseClient(BaseDatabaseClient):
    executable_name = "crash"

    def runshell(self, parameters):
        raise NotImplementedError()



import logging
import re
from collections.abc import Mapping
from itertools import tee

from crate.client.converter import DefaultTypeConverter
from crate.client.cursor import Cursor
from crate.client.connection import Connection

from django.core.exceptions import ImproperlyConfigured

from django.db.backends.base.base import BaseDatabaseWrapper
from django.utils.regex_helper import _lazy_re_compile

from .client import DatabaseClient
from .creation import DatabaseCreation
from .features import DatabaseFeatures
from .introspection import DatabaseIntrospection
from .operations import DatabaseOperations
from .schema import DatabaseSchemaEditor


def _get_varchar_column(data):
    if data["max_length"] is None:
        return "varchar"
    return "varchar(%(max_length)s)" % data


class E(Exception):
    pass


class CrateError:
    DataError = E
    OperationalError = E
    IntegrityError = E
    InternalError = E
    ProgrammingError = E
    NotSupportedError = E
    DatabaseError = E
    InterfaceError = E
    Error = E


class DatabaseWrapper(BaseDatabaseWrapper):
    vendor = "Crate.io"
    display_name = "CrateDB"
    Database = CrateError
    client_class = DatabaseClient
    creation_class = DatabaseCreation
    features_class = DatabaseFeatures
    introspection_class = DatabaseIntrospection
    ops_class = DatabaseOperations
    SchemaEditorClass = DatabaseSchemaEditor

    CRATE_SQL_SERIAL = "INTEGER GENERATED ALWAYS AS CAST((random() * 1.0E9) AS integer)"
    data_types = {
        # todo pgdiff - doc
        "AutoField": CRATE_SQL_SERIAL,
        "BigAutoField": CRATE_SQL_SERIAL,
        "SmallAutoField": CRATE_SQL_SERIAL,
        "BinaryField": "bytea",
        "BooleanField": "boolean",
        "CharField": _get_varchar_column,
        "DateField": "timestamp with time zone",
        "DateTimeField": "timestamp with time zone",
        "DecimalField": "numeric(%(max_digits)s, %(decimal_places)s)",
        "DurationField": "interval",
        "FileField": "varchar(%(max_length)s)",
        "FilePathField": "varchar(%(max_length)s)",
        "FloatField": "double precision",
        "IntegerField": "integer",
        "BigIntegerField": "bigint",
        "IPAddressField": "IP",
        "GenericIPAddressField": "inet",
        "JSONField": "object",
        "OneToOneField": "integer",
        "PositiveBigIntegerField": "bigint",
        "PositiveIntegerField": "integer",
        "PositiveSmallIntegerField": "smallint",
        "SlugField": "varchar(%(max_length)s)",
        "SmallIntegerField": "smallint",
        "TextField": "text",
        "TimeField": "time",
        "UUIDField": "text",
        "ObjectField": "OBJECT",
    }

    operators = {
        "exact": "= %s",
        "iexact": "= UPPER(%s)",
        "contains": "LIKE %s",
        "icontains": "LIKE UPPER(%s)",
        "regex": "~ %s",
        "iregex": "~* %s",
        "gt": "> %s",
        "gte": ">= %s",
        "lt": "< %s",
        "lte": "<= %s",
        "startswith": "LIKE %s",
        "endswith": "LIKE %s",
        "istartswith": "LIKE UPPER(%s)",
        "iendswith": "LIKE UPPER(%s)",
    }

    def rollback(self):
        return

    def savepoint(self):
        return

    def commit(self):
        return

    def _set_autocommit(self, autocommit):
        with self.wrap_database_errors:
            self.connection.autocommit = False  # Forcibly set autocommit to False.

    def get_connection_params(self):
        conn_params = dict(servers=self.settings_dict.get("SERVERS"))

        if self.settings_dict["HOST"]:
            conn_params["servers"] = [self.settings_dict["HOST"]]

        if self.settings_dict.get("PORT") or self.settings_dict.get("HOST"):
            raise ImproperlyConfigured(
                "Do not use 'PORT' nor 'HOST' in settings.databases, user 'SERVERS'"
            )
        return conn_params

    def get_new_connection(self, conn_params):
        return Connection(**conn_params)

    def create_cursor(self, name=None):
        return CrateDBCursorWrapper(self.connection, DefaultTypeConverter())


FORMAT_QMARK_REGEX = _lazy_re_compile(r"(?<!%)%s")


def aggressively_refresh():
    """
    Runs a refresh table statement on update statements
    """

    def deco(f):
        def wrapper(*args, **kwargs):
            func = f(*args, **kwargs)

            query = args[1].lower()
            match = re.search(r"update\s+([^\s;]+)", query, re.IGNORECASE)
            if match:
                table_name = match.group(1)
                return args[0].execute(f"refresh table {table_name}", None)
            return func
        return wrapper
    return deco


# Inspired by SQLITE driver
class CrateDBCursorWrapper(Cursor):
    """
    Django uses the "format" and "pyformat" styles, but CrateDB uses '?' question mark.

    This wrapper performs the following conversions:

    - "format" style to "qmark" style
    - "pyformat" style to "named" style

    In both cases, if you want to use a literal "%s", you'll need to use "%%s".
    """

    # todo pgdiff
    # @aggressively_refresh()
    def execute(self, query, params=None) -> None:
        if params is None:
            return super().execute(query)

        # Extract names if params is a mapping, i.e. "pyformat" style is used.
        param_names = list(params) if isinstance(params, Mapping) else None
        query = self.convert_query(query, param_names=param_names)
        logging.info(f"sent query: {query}, {params}")
        return super().execute(query, params)

    def executemany(self, query, param_list) -> int | list | None:
        # Extract names if params is a mapping, i.e. "pyformat" style is used.
        # Peek carefully as a generator can be passed instead of a list/tuple.
        peekable, param_list = tee(iter(param_list))
        if (params := next(peekable, None)) and isinstance(params, Mapping):
            param_names = list(params)
        else:
            param_names = None

        query = self.convert_query(query, param_names=param_names)
        logging.info(f"sent query: {query}")
        return super().executemany(query, param_list)

    def convert_query(self, query, *, param_names=None) -> str:
        if param_names is None:
            # Convert from "format" style to "qmark" style.
            # todo pgdiff
            return FORMAT_QMARK_REGEX.sub("?", query).replace("%%", "%")
        else:
            # Convert from "pyformat" style to "named" style.
            return query % {name: f":{name}" for name in param_names}



from django.db.backends.base.operations import BaseDatabaseOperations


class DatabaseOperations(BaseDatabaseOperations):
    compiler_module = "cratedb_django.compiler"

    def quote_name(self, name) -> str:
        if name.startswith('"') and name.endswith('"'):
            return name  # Quoting once is enough.
        return f'"{name}"'

    def sql_flush(
        self, style, tables, *, reset_sequences=False, allow_cascade=False
    ) -> list[str]:
        return [f"DELETE FROM {table}" for table in tables]

    def return_insert_columns(self, fields):
        """Returns the 'RETURNING...' part of the INSERT statement."""

        # We are supposed to return a (string) sql part with parameter binding and a (list) with
        # parameters, for example: ('returning ?, ?, ?', ['id1', 'id2', 'id'])
        # So the parameter binding happens at the cursor level.
        #
        # There is a bug that does not allow us to bind into the returning part of the query
        # See: https://github.com/crate/crate/issues/17813
        # until that is fixed we set everything in the sql part,
        # for example, ('returning id1, id2, id3', [])

        columns = ",".join(column.name for column in fields)
        return [f"returning {columns}", []]
